// ======================== Certis (C++) Libs ================================
//       LevelSets Library
// ===========================================================================
// Copyright (C) 2000-2007 Certis/Ecole des Ponts and Contributors
// For detailed information: http://www.enpc.fr/certis/software
// ===========================================================================

#ifndef _CLLEVELSETSFASTMARCHING_H
#define _CLLEVELSETSFASTMARCHING_H

#include <algorithm>
#include <limits>
#include <list>
#include <Imagine/Images.h>
#include "PriorityQueue.h"

#define EPS 1e-6


namespace Imagine {

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Enumeration pour l'etat des points
	typedef enum { eAlive=0, eTrial=1, eFar=2, eForbidden=3 } eState;	

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Classe contenant les coordonnees d'un point et la valeur de la fonction distance en ce point
	template <typename T, int dim>
	struct CoordsVal : public Coords<dim> {
		T val;
		CoordsVal(const Coords<dim> &p, T _val) : Coords<dim>(p), val(_val) {}

		bool operator < (const CoordsVal &a) const {
			return (val < a.val);
		}
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Classe pour le fast marching    
	template <typename T, int dim, int sign = +1>
	class FastMarching {
	private:
		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Types
		typedef PriorityQueue< CoordsVal<T,dim> > CoordsQueue;
		typedef PriorityQueueNode< CoordsVal<T,dim> > QueueNode;
		typedef std::list< Coords<dim> > CoordsList;
		typedef typename CoordsList::iterator CoordsListIterator;
		typedef typename CoordsList::const_iterator const_CoordsListIterator;

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Membres
	protected:
		Image<T,dim> &I;		  // Fonction distance
	private:
		Image<unsigned char,dim> state; // Etat des points
		Image<QueueNode *,dim> nodes; // Tableau des noeuds de la queue : necessaire pour modifier la priorite d'un point de la queue
		CoordsQueue trial_queue;  // Queue des points trial
		CoordsList alive_list;    // Liste des points alive
		T limit;				  // Condition d'arret du fast marching

	public:

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Constructeur / destructeur
		FastMarching(Image<T,dim> &_I) : I(_I), state(I.sizes()), nodes(I.sizes()) {
			init();
		}
		virtual ~FastMarching() {}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Iterateurs sur la liste de points alive
		inline const_CoordsListIterator begin() const { return alive_list.begin(); }
		inline const_CoordsListIterator end() const { return alive_list.end(); }

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Acces a l'etat d'un point
		inline eState get_state(const Coords<dim> &p) { return eState(state(p)); }
		inline eState get_state(size_t n) { return eState(state[n]); }

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Initialisation
		void init() {
			trial_queue.clear();
			alive_list.clear();
			state.fill(eFar);
			nodes.fill(0);
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Ajout d'un point trial a la queue
		inline void add_trial_point(const Coords<dim> &p) {
			size_t n = I.offset(p);
			if (state[n] == eFar) {
				state[n] = eTrial;
				nodes[n] = trial_queue.push( CoordsVal<T,dim>(p,sign*I[n]) );
			}
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Ajout d'un point alive a la liste
		inline void add_alive_point(const Coords<dim> &p) {
			size_t n = I.offset(p);
			if (state[n] == eFar) {
				state[n] = eAlive;
				alive_list.push_back(p);
			}
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Ajout d'un point interdit
		inline void add_forbidden_point(const Coords<dim> &p) {
			size_t n = I.offset(p);
			if (state[n] == eFar) {
				state[n] = eForbidden;
			}
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Generation des points trial a partir des point alive
		void init_trial_from_alive() {

			// On lit la liste de points alive et on ajoute les voisins a la queue des points trial
			for (const_CoordsListIterator p=begin();p!=end();++p) {
				for (int i=0;i<dim;i++) {
					if ((*p)[i]>0) {
						Coords<dim> pp(*p);
						pp[i]--;					
						add_trial_point(pp);
					}
					if ((*p)[i]<I.size(i)-1) {
						Coords<dim> pp(*p);
						pp[i]++;
						add_trial_point(pp);
					}
				}
			}
		}


		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Lancement du fast marching
		void run() {
			run( (sign>0) ? std::numeric_limits<T>::max() : -std::numeric_limits<T>::max() );
		}

		void run(T _limit) {
			// On memorise l'etendue du fast marching
			limit = _limit;

			// Tant qu'il y a des points trial dans la queue
			while(!trial_queue.empty()) {

				// On cherche le point trial de valeur la plus faible
				CoordsVal<T,dim> pt = trial_queue.top();

				// Si on a atteint la limite
				if (pt.val>=sign*limit) {
					// On vide la queue et on marque tous ses points comme far
					while (!trial_queue.empty()) {
						pt = trial_queue.top();
						trial_queue.pop();
						size_t n = I.offset(pt);
						state[n] = eFar;
						nodes[n] = 0;
						I[n] = limit;
					} 
					return;
				}

				// On enleve le point de la queue et on l'ajoute a la liste des points alive
				trial_queue.pop();
				size_t n = I.offset(pt);
				state[n] = eAlive;
				nodes[n] = 0;
				alive_list.push_back(pt);

				// On determine les voisins du point et on les met a jour
				for (int i=0;i<dim;i++) {
					if (pt[i]>0) {
						Coords<dim> pt2(pt);
						pt2[i]--;
						__update_point(pt2);
					}
					if (pt[i]<I.size(i)-1) {
						Coords<dim> pt2(pt);
						pt2[i]++;
						__update_point(pt2);
					}
				}
			}
		}


	private:
		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Mise a jour d'un point
		void __update_point(const Coords<dim> &p) {
			size_t n = I.offset(p);
			eState st = (eState)state[n];
			if (st == eFar) {
				// Si c'etait un point far, on le rend trial, on calcule sa valeur et on l'ajoute a la queue
				state[n] = eTrial;
				T val = _update_value(p);
				I[n] = sign*val;
				nodes[n] = trial_queue.push(CoordsVal<T,dim>(p,val));
			}
			else if (st == eTrial) {
				// Si c'etait deja un point trial, on recalcule sa valeur
				// Si la nouvelle valeur est inferieure, on l'accepte et on fait remonter le point dans la queue
				T val = _update_value(p);
				if (val<sign*I[n]) {
					I[n] = sign*val;
					trial_queue.increase(nodes[n],CoordsVal<T,dim>(p,val));
				}
			}
		}

	protected:
		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Acces a un point
		inline T _get_value(const Coords<dim> &p) const {
			size_t n = I.offset(p);
			if (state[n] <= eTrial) return sign*I[n];
			return std::numeric_limits<T>::max();
		}


		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Calcul de la valeur d'un point : a definir dans les classes derivees
		virtual T _update_value(const Coords<dim> &p) const = 0;


		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Resolution d'un trinome du second degre
		bool _solve_trinome(const T a, const T b, const T c, T &sol_max, T &sol_min) const {
			const T delta = b*b - 4*a*c;
			if (delta < 0) return false;
			const T sqrtDelta = std::sqrt(delta);
			sol_max = (- b + sqrtDelta) / a / 2;
			sol_min = (- b - sqrtDelta) / a / 2;
			return true;
		}
	};


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Equation iconale en 2D
	template <typename T, int dim, int sign = +1>
	class Eikonal : public FastMarching<T,dim,sign> {

	public:
		Eikonal(Image<T,dim> &_I, const Image<T,dim>& _speed) : FastMarching<T,dim,sign>(_I), speed(&_speed) {}
		Eikonal(Image<T,dim> &_I) : FastMarching<T,dim,sign>(_I), speed(0) {}
		virtual ~Eikonal() {}

	protected:
		const Image<T,dim> *speed;

		////////////////////////////////////////////////////////////////////////////////////////////////////////////     
		// Mise a jour d'un point
		virtual T _update_value(const Coords<dim> &p) const {

			// On prend le minimum de chaque paire de voisins
			T neighb[dim];
			for (int i=0;i<dim;i++) {
				Coords<dim> pp(p), np(p);
				pp[i]++; np[i]--;
				neighb[i] = (p[i]==0) ? this->_get_value(pp) : (p[i]==this->I.size(i)-1) ? this->_get_value(np) : std::min( this->_get_value(pp), this->_get_value(np) );
			}
			return _solve(neighb, speed ? (*speed)(p) : T(1));
		}


		//////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Resolution du trinome
		T _solve(T val[dim], T speed) const {

			// On reordonne les valeurs
			std::sort(val,val+dim);

			T b[dim], c[dim];
			b[0] = -val[0];
			c[0] = val[0]*val[0]-speed*speed;
			for (int i=1;i<dim;i++) {
				b[i] = b[i-1] - val[i];
				c[i] = c[i-1] + val[i]*val[i];
			}

			for (int i=dim-1;i>=1;i--) {
				T sol_max, sol_min;
				if (val[i]<std::numeric_limits<T>::max() && this->_solve_trinome(T(i+1), 2*b[i], c[i], sol_max, sol_min) && sol_max+EPS>=val[i]) return sol_max;
			}

			return val[0]+speed;
		}
	};

#undef EPS


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Function to transform a binary mask into a signed distance function using Fast Marching
	template <typename T, int dim>
	void mask_to_distance_function(Image<T,dim> &I, T thickness = 0) {

		Eikonal<T,dim,+1> eikonal1(I);
		Eikonal<T,dim,-1> eikonal2(I);

		const T initial_value = (thickness == 0) ? T(0.5) : thickness;
		for (typename Image<T,dim>::iterator i = I.begin() ; i != I.end() ; ++i) {
			*i = (*i > 0) ? -initial_value : initial_value;
		}

		for (Coords_iterator<dim> p = I.coords_begin() ; p != I.coords_end() ; ++p) {
			bool border = false;
			T I0 = I(*p);
			for (int i=0;i<dim;i++) {
				if ((*p)[i]>0) {
					Coords<dim> pp(*p);
					pp[i]--;
					if (I(pp)*I0 < 0) {
						border = true;
						break;
					}
				}
				if ((*p)[i]<I.size(i)-1) {
					Coords<dim> pp(*p);
					pp[i]++;
					if (I(pp)*I0 < 0) {
						border = true;
						break;
					}
				}
			}

			if (I0 > 0) {
				eikonal2.add_forbidden_point(*p);
				if (border)	{
					I(*p) = T(0.5);
					eikonal1.add_trial_point(*p);
				}
			}
			else {
				eikonal1.add_forbidden_point(*p);
				if (border)	{
					I(*p) = T(-0.5);
					eikonal2.add_trial_point(*p);
				}
			}
		}

		if (thickness == 0) {
			eikonal2.run();
			eikonal1.run();
		}
		else {
			eikonal2.run(-thickness);
			eikonal1.run(thickness);
		}
	}
}

#endif
