// ======================== Certis (C++) Libs ================================
//       LevelSets Library
// ===========================================================================
// Copyright (C) 2000-2007 Certis/Ecole des Ponts and Contributors
// For detailed information: http://www.enpc.fr/certis/software
// ===========================================================================

#ifndef _CLLEVELSETSLEVELSET_H
#define _CLLEVELSETSLEVELSET_H

#define _USE_MATH_DEFINES
#include <cmath>
#include <cstring>
#include <Imagine/Images.h>

#include "Fluxes.h"
#include "Approximators.h"
#include "Integrators.h"
#include "FastMarching.h"
#include "IO.h"

namespace Imagine {

	// Input iterator on a filtered range of coordinates
	template <int dim, class Predicate> class Filtered_coords_iterator : public Coords_iterator<dim> {
	protected:
		const Predicate& pred;
	public:
		Filtered_coords_iterator(const Coords<dim>& _a, const Coords<dim>& _b, const Predicate& _pred) : Coords_iterator<dim>(_a,_b), pred(_pred) {}
		Filtered_coords_iterator(const Predicate& _pred) : Coords_iterator<dim>(), pred(_pred) {}

		inline Filtered_coords_iterator& operator = (const Filtered_coords_iterator& other) {
			Coords_iterator<dim>::operator = (other);
			pred = other.pred;
			return *this;
		}	

		inline bool operator == (const Filtered_coords_iterator& other) const {
			return Coords_iterator<dim>::operator == (other);
		}

		bool operator != (const Filtered_coords_iterator& other) const {
			return Coords_iterator<dim>::operator != (other);
		}

		inline Filtered_coords_iterator& operator ++ () {
			do {
				Coords_iterator<dim>::operator ++ ();
			} while (!this->_stop && !pred(this->_pos));
			return *this;
		}

		inline Filtered_coords_iterator operator ++(int) {
			Filtered_coords_iterator tmp(*this);
			++(*this);
			return tmp;
		}
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Classe NarrowBand
	template <typename T, int dim> 
	class NarrowBand : public Image<bool,dim> {

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Membres
	protected:
		Image<T,dim>& I;				// Image
		Image<T,dim> prev;				// Fonction distance de la bande apres la derniere reinitialisation
		Eikonal<T,dim,+1> reinit1;		// Pour la reinitialisation par fast marching
		Eikonal<T,dim,-1> reinit2;

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Constructeur, destructeur et recopie
	public:
		NarrowBand(Image<T,dim>& _I) : Image<bool,dim>(_I.sizes()), I(_I), prev(_I.sizes()), reinit1(_I), reinit2(_I) {}

		// Iterator on the narrow band
		typedef Filtered_coords_iterator<dim, Image<bool,dim> > iterator;
		typedef iterator const_iterator;
		iterator begin() const { return iterator( Coords<dim>(0), this->sizes() - Coords<dim>(1), *this ); }
		iterator end() const { return iterator( *this ); }


		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Verifie si il faut reinitialiser la bande
		bool must_reinit(T threshold) const {

			for (iterator p = begin() ; p != end() ; ++p) {

				const size_t o = I.offset(*p);
				const T v = I[o];
				const T w = prev[o];

				if (w > threshold  && v <= 0) return true;
				if (w < -threshold && v >= 0) return true;
			}

			return false;
		}


		//////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//// Extension d'une donnee perpendiculairement a la surface
		//      template <class approximator>
		//void ExtendData(const Image<T,dim> &u, int iter = 20, T dt = 0.4) const {

		//	const Image<T,dim> tmp(u.dims());
		//	for (int t=0;t<iter;t++) {
		//		_MapBand(mask,dim,p) tmp(*p) = Extension<approximator>(*this,u,*p);
		//		_MapBand(mask,dim,p) u(*p) += dt * tmp(*p);
		//	}
		//}


		//////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//// Moyenne d'une quantite sur la surface
		//T BandAverage(const Image<T,dim> &u, const T neighb = 2.0, const int k = 0) const {
		//	T num = 0, den = 0;

		//	_const_MapBand(mask,dim,p) {
		//		const T _phi = (*this)(*p);
		//		T phix,phiy,phiz;
		//		Gradient3D(*this,p->x,p->y,p->z,phix,phiy,phiz);
		//		const T phig = ::sqrt(phix*phix+phiy*phiy+phiz*phiz);
		//		if (fabs(_phi) < neighb) {
		//			const T tmp = phig * (1 + cos(T(M_PI) * _phi / neighb)) / neighb / 2;
		//			den += tmp;
		//			num += tmp * u(vox->x,vox->y,vox->z,k);
		//		}
		//	}

		//	if (den < EPS) return 0;
		//	return num / den;
		//}


		//////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//// Integrale d'une quantite sur la surface
		//T BandIntegral(const Image<T,dim> &u, const T neighb = 2.0, const int k = 0) const {
		//	T sum = 0;
		//	_const_MapBand(mask,dim,p) {
		//		const T _phi = (*this)(*p);
		//		T phix,phiy,phiz;
		//		Gradient3D(*this,p->x,p->y,p->z,phix,phiy,phiz);
		//		const T phig = ::sqrt(phix*phix+phiy*phiy+phiz*phiz);
		//		if (fabs(_phi) < neighb) sum += u(vox->x,vox->y,vox->z,k) * phig * (1 + ::cos(M_PI * _phi / neighb)) / neighb / 2;
		//	}
		//	return sum;
		//}


		//////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//// Aire d'une surface
		//T Area(const T neighb = 2.0) const {
		//	T sum = 0;
		//	_const_MapBand(mask,dim,p) {
		//		const T _phi = (*this)(*p);
		//		T phix,phiy,phiz;
		//		Gradient3D(*this,p->x,p->y,p->z,phix,phiy,phiz);
		//		const T phig = ::sqrt(phix*phix+phiy*phiy+phiz*phiz);
		//		if (fabs(_phi) < neighb) sum += phig * (1 + cos(M_PI * _phi / neighb)) / neighb / 2;
		//	}
		//	return sum;
		//}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Ajout d'un point dans la bande
		void insert(const Coords<dim> &p) {
			(*this)(p) = true;
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Vidage de la bande
		void clear() {
			this->fill(false);
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Initialisation de la bande
		template <class approximator, class integrator>
		void init(T thickness, integrator& integr, int iter = 0, float dt = 0.4) {

			// Initialisation des objets de fast marching
			reinit1.init();
			reinit2.init();

			// On vide la bande
			clear();

			// On ajoute les points voisins du niveau 0 aux points alive
			for (Coords_iterator<dim> p = I.coords_begin() ; p != I.coords_end() ; ++p) {
				const T _phi = I(*p);
				Coords<dim> pp,mp;
				for(int i = 0 ; i < dim ; i++){
					pp = *p; mp = *p;
					if(mp[i] > 0) mp[i]--;
					if(pp[i] < I.size(i)-1) pp[i]++;
					if( _phi * I(pp) <= 0 || _phi * I(mp) <= 0){
						_add_alive_point(*p);
						break;
					}
				}
			}

			// On initialise les points trial a partir des points alive trouves precedemment
			reinit1.init_trial_from_alive();
			reinit2.init_trial_from_alive();

			// On ajoute dans la bande les points trial ou alive des deux cotes
			// On met la valeur des autres a 2 (condition aux bords correcte pour l'EDP)
			for (Coords_iterator<dim> p = I.coords_begin() ; p != I.coords_end() ; ++p) {
				const size_t o = I.offset(*p);
				const T _phi = I[o];
				if (_phi>0) {
					if (reinit1.get_state(o)==eFar)
						I[o] = T(2);
					else
						(*this)[o] = true;
				}
				else {
					if (reinit2.get_state(o)==eFar)
						I[o] = T(-2);
					else
						(*this)[o] = false;
				}
			}

			// EDP de reinitialisation
			_reinit_EDP<approximator>(integr,iter,dt);

			// Fast marching
			_reinit_fast_marching(thickness);

			// On met les points exterieurs a la valeur maximale du bon signe
			for (Coords_iterator<dim> p = I.coords_begin() ; p != I.coords_end() ; ++p) {
				const size_t o = I.offset(*p);
				const T _phi = I[o];
				if (_phi>0 && reinit1.get_state(o)==eFar) I[o] = thickness;
				else if (_phi<0 && reinit2.get_state(o)==eFar) I[o] = -thickness;
			}

			// On memorise le resultat de l'initialisation
			memcpy(prev.data(),I.data(),I.Array<T>::size()*sizeof(T));
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Reinitialisation de la bande
		template <class approximator, class integrator>
		void reinit(T thickness, integrator& integr, int iter = 2, T dt = 0.4) {

			// EDP de reinitialisation
			_reinit_EDP<approximator>(integr,iter,dt);

			// Initialisation des objets de fast marching
			reinit1.init();
			reinit2.init();

			// Les point alive sont les points voisins du niveau 0
			for (iterator p = begin() ; p != end() ; ++p) {
				const T _phi = I(*p);
				Coords<dim> pp,mp;
				for(int i = 0 ; i < dim ; i++){
					pp = *p; mp = *p;
					if (mp[i] > 0) mp[i]--;
					if (pp[i] < I.size(i)-1) pp[i]++;
					if ( _phi * I(pp) <= 0 || _phi * I(mp) <= 0){
						_add_alive_point(*p);
						break;
					}
				}
			}

			// Les points trial sont les voisins des points alive
			reinit1.init_trial_from_alive();
			reinit2.init_trial_from_alive();

			// On enleve de la bande les points qui ne sont ni trial ni alive des deux cotes
			// et on met leur valeur au max
			for (iterator p = begin() ; p != end() ; ++p) {
				const size_t o = I.offset(*p);
				const T _phi = I[o];
				if (_phi>0 && reinit1.get_state(o)==eFar) {
					I[o] = thickness;
					(*this)[o] = false;
				}
				else if (_phi<=0 && reinit2.get_state(o)==eFar) {
					I[o] = -thickness;
					(*this)[o] = false;
				}
			}

			// Fast marching
			_reinit_fast_marching(thickness);

			// On memorise le resultat de la reinitialisation
			memcpy(prev.data(),I.data(),I.Array<T>::size()*sizeof(T));
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Maintien de la fonction distance
		template <class approximator, class integrator>
		void maintain(T thickness, integrator& integr, int expand = 1, int iter = 1, T dt = 0.4) {
			_expand(expand);
			_reinit_EDP<approximator>(integr,iter,dt);
			_shrink(thickness);
		}

	protected:
		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Ajout d'un point alive
		void _add_alive_point(const Coords<dim>& p) {
			const T _phi = I(p);
			if (_phi>0) {
				reinit1.add_alive_point(p);
				reinit2.add_forbidden_point(p);
			}
			else {
				reinit1.add_forbidden_point(p);
				reinit2.add_alive_point(p);
			}	
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// EDP de reinitialisation
		template <class approximator, class integrator>
		void _reinit_EDP(integrator& integr, int iter = 2, T dt = 0.4) {
			for (int t=0;t<iter;t++) {
				do {
					for (iterator p = begin() ; p != end() ; ++p)
						integr(*p) = reinitialization<approximator>(I,*p);
				} while (!integr.step(begin(),end(),dt));
			}
		}


		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Reinitialisation par fast marching
		void _reinit_fast_marching(T thickness, T keep = 0) {

			// On realise les deux fast marching
			if (thickness>0) {
				reinit1.run(thickness);
				reinit2.run(-thickness);
			}
			else {
				// Fast marching dans tout le domaine
				reinit1.run();
				reinit2.run();
			}

			// On ajoute a la bande les points alive trouves au cours du fast marching
			typedef typename std::list< Coords<dim> >::const_iterator const_CoordsListIterator;
			if (keep==0) {
				for (const_CoordsListIterator p=reinit1.begin();p!=reinit1.end();++p) insert(*p);
				for (const_CoordsListIterator p=reinit2.begin();p!=reinit2.end();++p) insert(*p);
			}
			else {
				for (const_CoordsListIterator p=reinit1.begin();p!=reinit1.end();++p) if (I(*p)<keep) insert(*p);
				for (const_CoordsListIterator p=reinit2.begin();p!=reinit2.end();++p) if (I(*p)>-keep) insert(*p);
			}
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Dilatation de la bande
		void _expand(int expand) {

			// On effectue plusieurs fois une dilatation elementaire
			for (int k=0;k<expand;k++) {
				for (iterator p = begin() ; p != end() ; ++p) {
					Coords<dim> pp,mp;
					neighbour_coords(I,*p,pp,mp);
					for(Coords_iterator<dim> it(mp,pp) ; it != Coords_iterator<dim>() ; ++it) insert(*it);
				}
			}
		}

		////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Reduction de la bande
		void _shrink(T thickness) {

			// On parcourt les points de la bande
			for (iterator p = begin() ; p != end() ; ++p) {
				const int o = I.offset(*p);
				const T v = I[o];

				// Suivant la valeur de la fonction distance
				if (v > thickness) {
					// On enleve le point
					(*this)[o] = false;

					// On donne a la fonction de niveau la valeur maximum
					prev[o] = I[o] = thickness;
				}
				else if (v < -thickness){
					// On enleve le point
					(*this)[o] = false;

					// On donne a la fonction de niveau la valeur minimum
					prev[o] = I[o] = - thickness;
				}
				else {
					// Le point est dans la bande, on memorise la valeur de la fonction de niveau
					prev[o] = v;
				}
			}
		}
	};
 }

#endif
