// ===========================================================================
// Imagine++ Libraries
// Copyright (C) Imagine
// For detailed information: http://imagine.enpc.fr/software
// ===========================================================================

#ifndef sqr
#define sqr(x) ((x)*(x))
#endif

namespace Imagine {
/// \addtogroup LevelSets
	/// @{

	

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Differences centrees du 2eme ordre
	class Centered2 {
	public:
		template<typename T,int dim> static T u_i(const Image<T,dim>& u, const Coords<dim>& p, int i) {
			assert(i>=0 && i<dim);
			const int o = u.offset(p);
			return ( ( ( p[i] < u.size(i)-1 ) ? u[ o + u.stride(i) ] : u[o] ) - ( ( p[i] > 0 ) ? u[ o - u.stride(i) ] : u[o] ) ) / T(2);
		}


		template<typename T, int dim> static T u_pi(const Image<T,dim> &u, const Coords<dim> &p, int i) {
			return u_i(u,p,i);
		}


		template<typename T, int dim> static T u_mi(const Image<T,dim> &u, const Coords<dim> &p, int i) {
			return u_i(u,p,i);
		}
	};


	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Upwind
	class Upwind {
	public:
		template<typename T, int dim> static T u_pi(const Image<T,dim>& u, const Coords<dim>& p, int i) {
			assert(i>=0 && i<dim);
			if (p[i] < u.size(i)-1) {
				const int o = u.offset(p);
				return u[ o + u.stride(i) ] - u[o];
			}
			return 0;
		}


		template<typename T, int dim> static T u_mi(const Image<T,dim>& u, const Coords<dim>& p, int i) {
			assert(i>=0 && i<dim);
			if (p[i] > 0) {
				const int o = u.offset(p);
				return u[o] - u[ o - u.stride(i) ];
			}
			return 0;
		}
	};


#define _CLLEVELSETAPPROXIMATORS_EPS 1e-12

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Weno3
	class Weno3 {
	private:
		template<typename T> static T combine(const T v1, const T v2, const T v3) {
			T s = T(( T( _CLLEVELSETAPPROXIMATORS_EPS) + sqr(v2 - v1) ) / ( T( _CLLEVELSETAPPROXIMATORS_EPS) + sqr(v3 - v2) ));
			s = T(1) / (T(1)+T(2)*s*s);
			return (v2+v3 - s * (v1 - 2*v2 + v3)) / T(2);
		}

	public:
		template<typename T,int dim> static T u_pi(const Image<T,dim>& u, const Coords<dim>& p, int i) {
			assert(i>=0 && i<dim);
			const size_t stride = u.stride(i);
			const size_t o = u.offset(p);
			const T u1 = u[o];
			const T u0 = (p[i] > 0) ? u[o - stride] : u1;
			const T u2 = (p[i] < u.size(i)-1) ? u[o + stride] : u1;
			const T u3 = (p[i] < u.size(i)-2) ? u[o + 2*stride] : u2;
			return combine(u3-u2,u2-u1,u1-u0);
		}


		template<typename T,int dim> static T u_mi(const Image<T,dim> &u, const Coords<dim> &p, int i) {
			assert(i>=0 && i<dim);
			const size_t stride = u.stride(i);
			const size_t o = u.offset(p);
			const T u2 = u[o];
			const T u1 = (p[i] > 0) ? u[o - stride] : u2;
			const T u0 = (p[i] > 1) ? u[o - 2*stride] : u1;
			const T u3 = (p[i] < u.size(i)-1) ? u[o + stride] : u2;
			return combine(u1-u0,u2-u1,u3-u2);
		}
	};


	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Weno5
	class Weno5 {
	public:
		template<typename T> static T combine(const T v1, const T v2, const T v3, const T v4, const T v5) {

			T s1 = T(13) * sqr(v1 - T(2)*v2 + v3) / T(12) + sqr(v1 - T(4)*v2 + T(3)*v3) / T(4);
			T s2 = T(13) * sqr(v2 - T(2)*v3 + v4) / T(12) + sqr(v2 - v4) / T(4);
			T s3 = T(13) * sqr(v3 - T(2)*v4 + v5) / T(12) + sqr(T(3)*v3 - T(4)*v4 + v5) / T(4);

			s1 = T(1) / sqr(T( _CLLEVELSETAPPROXIMATORS_EPS) + s1);
			s2 = T(6) / sqr(T( _CLLEVELSETAPPROXIMATORS_EPS) + s2);
			s3 = T(3) / sqr(T( _CLLEVELSETAPPROXIMATORS_EPS) + s3);

			return ( s1 * (T(2)*v1 - T(7)*v2 + T(11)*v3) + s2 * (-v2 + T(5)*v3 + T(2)*v4) + s3 * (T(2)*v3 + T(5)*v4 - v5) ) / T(6) / (s1+s2+s3);
		}


		template<typename T,int dim> static T u_pi(const Image<T,dim>& u, const Coords<dim>& p, int i) {
			assert(i>=0 && i<dim);
			const int stride = u.stride(i);
			const int o = u.offset(p);
			const T u2 = u[o];
			const T u1 = (p[i] > 0) ? u[o - stride] : u2;
			const T u0 = (p[i] > 1) ? u[o - 2*stride] : u1;
			const T u3 = (p[i] < u.size(i)-1) ? u[o + stride] : u2;
			const T u4 = (p[i] < u.size(i)-2) ? u[o + 2*stride] : u3;
			const T u5 = (p[i] < u.size(i)-3) ? u[o + 3*stride] : u4;
			return combine(u5-u4,u4-u3,u3-u2,u2-u1,u1-u0);
		}


		template<typename T,int dim> static T u_mi(const Image<T,dim>& u, const Coords<dim>& p, int i) {
			assert(i>=0 && i<dim);
			const int stride = u.stride(i);
			const int o = u.offset(p);
			const T u3 = u[o];
			const T u2 = (p[i] > 0) ? u[o - stride] : u3;
			const T u1 = (p[i] > 1) ? u[o - 2*stride] : u2;
			const T u0 = (p[i] > 2) ? u[o - 3*stride] : u1;
			const T u4 = (p[i] < u.size(i)-1) ? u[o + stride] : u3;
			const T u5 = (p[i] < u.size(i)-2) ? u[o + 2*stride] : u4;
			return combine(u1-u0,u2-u1,u3-u2,u4-u3,u5-u4);
		}
	};

	///@}
 }
