// ===========================================================================
// Imagine++ Libraries
// Copyright (C) Imagine
// For detailed information: http://imagine.enpc.fr/software
// ===========================================================================

namespace Imagine {
/// \addtogroup LevelSets
	/// @{

	
	// Structure decrivant un cas de marching squares
	struct MsCase {
		int edges[5];
	};

	// Correspondance entre les aretes et les points
	const int MS_EDGES_TABLE[4][2] = {
		{0,1}, {1,2}, {3,2}, {0,3}
	};

	// Table des cas
	const MsCase MS_CASES_TABLE[] = {
		{{-1, -1, -1, -1, -1}},
		{{0, 3, -1, -1, -1}},
		{{1, 0, -1, -1, -1}},
		{{1, 3, -1, -1, -1}},
		{{2, 1, -1, -1, -1}},
		{{2, 1, 0, 3, -1}},
		{{2, 0, -1, -1, -1}},
		{{2, 3, -1, -1, -1}},
		{{3, 2, -1, -1, -1}},
		{{0, 2, -1, -1, -1}},
		{{3, 2, 1, 0, -1}},
		{{1, 2, -1, -1, -1}},
		{{3, 1, -1, -1, -1}},
		{{0, 1, -1, -1, -1}},
		{{3, 0, -1, -1, -1}},
		{{-1, -1, -1, -1, -1}}
	};


	template <typename T, class OutputIterator1, class OutputIterator2>
	void marchingSquares(const Image<T,2>& phi, T value, OutputIterator1 points, OutputIterator2 normals, const Affine<T,2>& affine = Affine<T,2>::Identity())
	{
		// Matrice de transformation de la normale et calcul du gradient de phi
		FMatrix<T,2,2> affineN = transpose(affine.matrix());
		const T sigma = 0.5;
		Image<T,2>
			phix = deriche(phi,sigma,1,0),
			phiy = deriche(phi,sigma,1,1);

		// Valeurs aux coins du carre
		T v[4];

		// Increments pour acceder aux voisins
		const size_t dx = phi.stride(0);
		const size_t dy = phi.stride(1);

		// On parcourt l'image
		for (int y=0;y<phi.height()-1;y++) for (int x=0;x<phi.width()-1;x++) {

			// On calcule les valeurs aux 8 coins du cube
			size_t n = phi.offset(x,y);
			v[0] = phi[n];
			v[1] = phi[n+dx];
			v[2] = phi[n+dx+dy];
			v[3] = phi[n+dy];

			// On calcule le numero du cas
			int index = 0;
			for (int i=0;i<4;i++) if (v[i]<value) index |= 1<<i;

			// On traite separement le cas 0 et 15 tres frequent
			if (index == 0 || index == 15) continue;

			// On va chercher le cas dans la table
			const int *cas = MS_CASES_TABLE[index].edges;

			// Pour chaque segment du cas
			for (;cas[0]>-1;cas+=2) {
				// Pour chaque point du segment
				for (int edge=0;edge<2;edge++) {
					const int edge_id = cas[edge];
					const int *pts = MS_EDGES_TABLE[edge_id];
					// Indice entre 0 et 3 des points a interpoler 
					const int pt1 = pts[0];
					const int pt2 = pts[1];
					// Coordonnee barycentrique
					const T f = (value - v[pt1]) / (v[pt2] - v[pt1]);
					// Coordonnees du point
					T xx,yy;
					switch (edge_id) {
						case 0 : xx = T(x) + f; yy = T(y); break;
						case 1 : xx = T(x) + 1; yy = T(y) + f; break;
						case 2 : xx = T(x) + f; yy = T(y) + 1; break;
						case 3 : xx = T(x); yy = T(y) + f; break;
						default : assert(false);
					}

					// On rajoute le point dans le tableau
					FVector<T,2> p,n;
					affine.apply(xx,yy,p.x(),p.y());
					T nx =(T) phix.interpolate(xx,yy), ny = (T)phiy.interpolate(xx,yy);
					n = affineN * FVector<T,2>(nx,ny);
					*points = p;
					*normals++ = normalized(n);
				}
			}
		}
	}
	///@}
}
